import graph_scheduler
import pytest

import modeci_mdf.mdf as mdf
from modeci_mdf.execution_engine import EvaluableGraph


@pytest.mark.parametrize(
    "function, value, result",
    [
        ("linear", None, 6),
        # execution is in favor of function not value
        ("linear", "2 * slope * variable0 + intercept", 6),
        (None, "2 * slope * variable0 + intercept", 8),
        (None, "slope + intercept", 6),
    ],
)
@pytest.mark.parametrize(
    "args",
    [
        {"slope": 2, "intercept": 4, "variable0": 1},
        {"slope": 2, "intercept": 4, "variable0": "input"},
        {"slope": 2, "intercept": 4, "variable0": "input"},
        {"slope": "2 * input", "intercept": 4, "variable0": "input"},
        {"slope": 2, "intercept": "2 * slope", "variable0": "input"},
        {"slope": "math.sqrt(4)", "intercept": 4, "variable0": 1},
        {"slope": "numpy.sqrt(4)", "intercept": 4, "variable0": 1},
    ],
)
def test_single_function_variations(create_model, args, function, value, result):
    m = create_model(
        [
            mdf.Node(
                id="N",
                input_ports=[mdf.InputPort(id="input")],
                functions=[
                    mdf.Function(id="f", args=args, function=function, value=value)
                ],
                output_ports=[mdf.OutputPort(id="output", value="f")],
            )
        ]
    )

    eg = EvaluableGraph(m.graphs[0])
    eg.evaluate(initializer={"input": 1})

    assert eg.enodes["N"].evaluable_outputs["output"].curr_value == result


@pytest.mark.parametrize(
    "node_specific, termination, result",
    [
        ({"A": mdf.Condition(type="Always")}, None, 1),
        (
            None,
            {
                graph_scheduler.TimeScale.ENVIRONMENT_STATE_UPDATE: mdf.Condition(
                    type="AfterNCalls", dependency="A", n=5
                )
            },
            5,
        ),
    ],
)
def test_condition_variations(create_model, node_specific, termination, result):
    A = mdf.Node(
        id="A",
        input_ports=[mdf.InputPort(id="A_input")],
        parameters=[
            mdf.Parameter(id="A_param", value="A_param + 1", default_initial_value=0)
        ],
        output_ports=[mdf.OutputPort(id="A_output", value="A_param")],
    )

    model = create_model(
        nodes=[A],
        conditions=mdf.ConditionSet(
            node_specific=node_specific, termination=termination
        ),
    )

    eg = EvaluableGraph(model.graphs[0])
    eg.evaluate(initializer={"A_input": 0})

    assert eg.enodes["A"].evaluable_outputs["A_output"].curr_value == result


def test_dependency_in_function_value(create_model):
    m = create_model(
        [
            mdf.Node(
                id="N",
                input_ports=[mdf.InputPort(id="input")],
                functions=[
                    mdf.Function(id="f", value="g"),
                    mdf.Function(id="g", value="1"),
                ],
                output_ports=[mdf.OutputPort(id="output", value="f")],
            )
        ]
    )

    eg = EvaluableGraph(m.graphs[0])
    eg.evaluate(initializer={"input": 1})


# NOTE: this is enabled by "don't include known args in required variable"
# but could not be tested until dependency checking for value
def test_available_arg_in_function_value(create_model):
    m = create_model(
        [
            mdf.Node(
                id="N",
                input_ports=[mdf.InputPort(id="input")],
                functions=[
                    mdf.Function(id="f", args={"g": 1}, value="g"),
                ],
                output_ports=[mdf.OutputPort(id="output", value="f")],
            )
        ]
    )

    eg = EvaluableGraph(m.graphs[0])
    eg.evaluate(initializer={"input": 1})



"""Import and export code for `ONNX <https://onnx.ai/>`_ models"""

from .importer import (
    onnx_to_mdf,
    find_subgraphs,
    convert_file,
    get_color_for_onnx_category,
    get_category_of_onnx_node,
)

from .exporter import mdf_to_onnx



import pytest
from modeci_mdf.execution_engine import get_required_variables_from_expression


@pytest.mark.parametrize(
    "expression, expected_variables",
    [
        (0, []),
        ((), []),
        ("", []),
        ("x", ["x"]),
        ("x.y", ["x"]),
        ("3*x + y", ["x", "y"]),
        ("x[y]", ["x", "y"]),
        ("x[y[z]]", ["x", "y", "z"]),
        ("x[y[z + 1]]", ["x", "y", "z"]),
        ("x[y] + z[a]", ["x", "y", "z", "a"]),
        ("x[y[z]] + a[b]", ["x", "y", "z", "a", "b"]),
        ("x[y[z[1]]] + a[0]", ["x", "y", "z", "a"]),
        ("x[y[z] + 1]", ["x", "y", "z"]),
        ("x[y[z[1 + a] + b[c]]] + d[0]", ["x", "y", "z", "a", "b", "c", "d"]),
        ("[x, 0, 1]", ["x"]),
    ],
)
def test_expression_parsing(expression, expected_variables):
    assert set(get_required_variables_from_expression(expression)) == set(
        expected_variables
    )
    assert set(get_required_variables_from_expression(f"[{expression}]")) == set(
        expected_variables
    )



import math

import numpy
import pytest

import modeci_mdf.functions.standard as stdf


@pytest.mark.parametrize(
    "name, parameters, expected_result",
    [
        ("linear", {"variable0": 1, "slope": 2, "intercept": 3}, 5),
        (
            "logistic",
            {"variable0": 1, "gain": 2, "bias": 3, "offset": 4},
            0.9820137900379085,
        ),
        (
            "exponential",
            {"variable0": 1, "scale": 2, "rate": 3, "bias": 4, "offset": 5},
            2198.266316856917,
        ),
        ("sin", {"variable0": math.pi / 2, "scale": 2}, 2.0),
        ("cos", {"variable0": math.pi, "scale": 2}, -2.0),
        (
            "MatMul",
            {"A": numpy.array([[1, 2], [3, 4]]), "B": numpy.array([[1, 2], [3, 4]])},
            numpy.array([[7, 10], [15, 22]]),
        ),
        ("Relu", {"A": 1}, 1),
        ("Relu", {"A": -1}, 0),
    ],
)
def test_std_functions(name, expected_result, parameters):
    try:
        assert stdf.mdf_functions[name]["function"](**parameters) == expected_result
    except ValueError:
        assert numpy.array_equal(
            stdf.mdf_functions[name]["function"](**parameters), expected_result
        )



"""
Test some individual ONNX operator calls. We should probably find a better way to
test these more exhaustively by hooking into ONNX's testing framework. For now,
I have hand coded some tests for operators that are used in the examples we have
worked on. There could be broken operators.
"""
import numpy as np
import modeci_mdf.functions.onnx as onnx_ops


def test_conv():
    """Test ONNX Conv function"""
    x = np.array(
        [
            [
                [
                    [0.0, 1.0, 2.0, 3.0, 4.0],  # (1, 1, 5, 5) input tensor
                    [5.0, 6.0, 7.0, 8.0, 9.0],
                    [10.0, 11.0, 12.0, 13.0, 14.0],
                    [15.0, 16.0, 17.0, 18.0, 19.0],
                    [20.0, 21.0, 22.0, 23.0, 24.0],
                ]
            ]
        ]
    ).astype(np.float32)
    W = np.array(
        [
            [
                [
                    [1.0, 1.0, 1.0],  # (1, 1, 3, 3) tensor for convolution weights
                    [1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0],
                ]
            ]
        ]
    ).astype(np.float32)

    out = onnx_ops.conv(x, W)


def test_pad():
    """Test ONNX Pad function"""
    x = np.zeros((3, 2))
    value = np.array(1.5)
    pads = np.array([0, 1, 0, 1]).astype(np.int64)

    out = onnx_ops.pad(x, pads, value, mode="constant")

    # Try attributes without keyword
    out2 = onnx_ops.pad(x, pads, value, "constant")

    assert np.all(out == out2)


def test_pad_diff_types():
    """Check if Pad can handle the case were a different type is passed to constant_value than the type of the data"""

    args = {
        "data": np.zeros((1, 48, 32, 32), dtype=np.float32),
        "pads": np.array([0, 0, 0, 0, 0, 0, 1, 1], dtype=np.int64),
        "constant_value": 1.5,
        "mode": "constant",
    }

    onnx_ops.pad(**args)


def test_unsqueeze():
    """Test ONNX unsqueeze function."""
    data = np.zeros((3, 2))
    axes = [0]

    out = onnx_ops.unsqueeze(data=data, axes=axes)

    assert out.ndim == 3
    assert out.shape == (1, 3, 2)


def test_mul():
    """Test element-wise tensor multiplication (Mul)"""
    A = np.ones((1, 3)) * 2.0
    B = np.ones((3, 1))
    assert np.allclose(A * B, onnx_ops.mul(A, B))

    A = 1
    B = 2
    assert np.allclose(A * B, onnx_ops.mul(A, B))


def test_constantofshape():
    """Test ConstantOfShape function."""
    out = onnx_ops.constantofshape(np.array([4, 4], dtype=np.int64), value=[0])
    assert np.allclose(out, np.zeros((4, 4), dtype=np.int64))


def test_concat():
    """Test ONNX Concat function. This is a variable number of inputs operator."""
    input = (np.ones(3), np.ones(3), np.ones(3))
    out = onnx_ops.concat(*input, axis=0)
    assert np.allclose(out, np.concatenate(input, axis=0))


def test_maxpool():
    """Test ONNX Concat function. This is a variable number of inputs operator."""
    out = onnx_ops.maxpool(
        np.ones((1, 3, 32, 32)).astype(np.float32), kernel_shape=[2, 2]
    )
    assert True



import graph_scheduler
import numpy as np
import pytest

import modeci_mdf.mdf as mdf
from modeci_mdf.execution_engine import EvaluableGraph


def test_execution_engine_main(tmpdir):

    import modeci_mdf.execution_engine
    from modelspec.utils import FORMAT_NUMPY, FORMAT_TENSORFLOW
    import numpy as np

    mdf_formats = ["json", "yaml"]
    array_formats = [FORMAT_NUMPY, FORMAT_TENSORFLOW]

    # For now, don't make tensorflow a requiremnt...
    try:
        import tensorflow
    except:
        array_formats = [FORMAT_NUMPY]

    for mdf_format in mdf_formats:
        for array_format in array_formats:

            eg = modeci_mdf.execution_engine.main(
                "examples/MDF/Simple.%s" % mdf_format, array_format=array_format
            )
            output = (
                eg.enodes["processing_node"].evaluable_outputs["output_1"].curr_value
            )
            assert output == 0.6016871801828567

            eg = modeci_mdf.execution_engine.main(
                "examples/MDF/ABCD.%s" % mdf_format, array_format=array_format
            )
            output = eg.enodes["D"].evaluable_outputs["out_port"].curr_value
            assert -1.7737500239216304 - output < 1e-9

            eg = modeci_mdf.execution_engine.main(
                "examples/MDF/Arrays.%s" % mdf_format, array_format=array_format
            )
            output = eg.enodes["middle_node"].evaluable_outputs["output_1"].curr_value
            assert output[0, 0] == 0.5
            assert output[1, 1] == 4

            eg = modeci_mdf.execution_engine.main(
                "examples/MDF/States.%s" % mdf_format, array_format=array_format
            )
            output = eg.enodes["counter_node"].evaluable_outputs["out_port"].curr_value
            assert output == 1
            output = eg.enodes["sine_node"].evaluable_outputs["out_port"].curr_value
            assert output == 0


def test_execution_engine_onnx(tmpdir):

    import modeci_mdf.execution_engine
    import numpy as np

    mdf_formats = ["json", "yaml"]
    from modelspec.utils import FORMAT_NUMPY

    array_format = FORMAT_NUMPY

    for mdf_format in mdf_formats:

        eg = modeci_mdf.execution_engine.main(
            "examples/ONNX/ab.%s" % mdf_format, array_format=array_format
        )
        output = eg.enodes["Mul_3"].evaluable_outputs["_4"].curr_value
        assert np.array_equal(output, np.full((2, 3), 5))


_abc_conditions_expected_output = [
    {"A"},
    {"A"},
    {"B"},
    {"A"},
    {"C"},
    {"A"},
    {"B"},
    {"A"},
    {"A"},
    {"C", "B"},
    {"A"},
]


@pytest.mark.parametrize(
    "fi, expected_output",
    [
        ("examples/MDF/abc_conditions.json", _abc_conditions_expected_output),
        ("examples/MDF/abc_conditions.yaml", _abc_conditions_expected_output),
    ],
)
def test_condition_scheduler_main(fi, expected_output):

    import modeci_mdf.execution_engine

    eg = modeci_mdf.execution_engine.main(fi)
    output = [{n.id for n in nodes} for nodes in eg.scheduler.execution_list[None]]
    assert output == expected_output

    assert eg.enodes["A"].evaluable_parameters["count_A"].curr_value == 7
    assert eg.enodes["B"].evaluable_parameters["count_B"].curr_value == 3


def test_nested_conditions(create_model):
    A = mdf.Node(
        id="A",
        input_ports=[mdf.InputPort(id="A_input")],
        parameters=[
            mdf.Parameter(id="A_param", value="A_param + 1", default_initial_value=0)
        ],
        output_ports=[mdf.OutputPort(id="A_output", value="A_param")],
    )
    B = mdf.Node(
        id="B",
        input_ports=[mdf.InputPort(id="B_input")],
        parameters=[
            mdf.Parameter(id="B_param", value="B_param + 1", default_initial_value=0)
        ],
        output_ports=[mdf.OutputPort(id="B_output", value="B_param")],
    )
    C = mdf.Node(
        id="C",
        input_ports=[mdf.InputPort(id="C_input")],
        parameters=[
            mdf.Parameter(id="C_param", value="C_param + 1", default_initial_value=0)
        ],
        output_ports=[mdf.OutputPort(id="C_output", value="C_param")],
    )

    m = create_model(
        nodes=[A, B, C],
        conditions=mdf.ConditionSet(
            node_specific={
                "B": mdf.Condition(
                    type="EveryNCalls", kwargs={"dependency": A, "n": 2}
                ),
                # condition is simple EveryNCalls(B, 2) within two Not conditions
                "C": mdf.Condition(
                    type="Not",
                    kwargs={
                        "condition": mdf.Condition(
                            type="Not",
                            kwargs={
                                "condition": mdf.Condition(
                                    type="EveryNCalls", kwargs={"dependency": B, "n": 2}
                                )
                            },
                        )
                    },
                ),
            },
            termination={
                graph_scheduler.TimeScale.ENVIRONMENT_STATE_UPDATE: mdf.Condition(
                    type="AfterNCalls", kwargs={"dependency": C, "n": 2}
                )
            },
        ),
    )

    m2 = mdf.Model.from_dict(m.to_dict())
    m3 = mdf.Model.from_dict(m2.to_dict())

    for model in [m, m2, m3]:
        eg = EvaluableGraph(model.graphs[0])
        eg.evaluate(initializer={"A_input": 0})

        assert eg.enodes["A"].evaluable_outputs["A_output"].curr_value == 8
        assert eg.enodes["B"].evaluable_outputs["B_output"].curr_value == 4
        assert eg.enodes["C"].evaluable_outputs["C_output"].curr_value == 2


@pytest.mark.skipif(
    graph_scheduler.__version__ < "1.1.0", reason="Threshold added in 1.1.0"
)
def test_threshold(create_model):
    A = mdf.Node(
        id="A",
        input_ports=[mdf.InputPort(id="A_input")],
        parameters=[
            mdf.Parameter(id="A_param", value="A_param + 1", default_initial_value=0)
        ],
        output_ports=[mdf.OutputPort(id="A_output", value="A_param")],
    )

    m = create_model(
        nodes=[A],
        conditions=mdf.ConditionSet(
            termination={
                graph_scheduler.TimeScale.ENVIRONMENT_STATE_UPDATE: mdf.Condition(
                    type="Threshold",
                    dependency=A,
                    parameter="A_param",
                    threshold=5,
                    comparator=">=",
                )
            },
        ),
    )

    m2 = mdf.Model.from_dict(m.to_dict())
    m3 = mdf.Model.from_dict(m2.to_dict())

    for model in [m, m2, m3]:
        eg = EvaluableGraph(model.graphs[0])
        eg.evaluate(initializer={"A_input": 0})

        assert eg.enodes["A"].evaluable_outputs["A_output"].curr_value == 5


@pytest.mark.skipif(
    graph_scheduler.__version__ < "1.1.0", reason="Threshold added in 1.1.0"
)
@pytest.mark.parametrize(
    "threshold_dependency, threshold_parameter",
    [
        ("A", "A_output"),
        ("A", "A_param_2"),
        ("B", "B_input"),
    ],
)
@pytest.mark.parametrize(
    "threshold_indices, result",
    [
        ((0,), np.array([5, 6])),
        ((1,), np.array([4, 5])),
    ],
)
def test_threshold_nonscalar_values(
    create_model, threshold_dependency, threshold_parameter, threshold_indices, result
):
    A = mdf.Node(
        id="A",
        input_ports=[mdf.InputPort(id="A_input")],
        parameters=[
            mdf.Parameter(id="A_param", value="A_param + 1", default_initial_value=0),
            mdf.Parameter(
                id="A_param_2",
                value="[A_param_2[0] + 1, A_param_2[1] + 1]",
                default_initial_value=np.array([0, 1]),
            ),
        ],
        output_ports=[mdf.OutputPort(id="A_output", value="A_param_2", shape=(2,))],
    )
    B = mdf.Node(
        id="B",
        input_ports=[mdf.InputPort(id="B_input", shape=(2,))],
        output_ports=[mdf.OutputPort(id="B_output", value="B_input", shape=(2,))],
    )

    m = create_model(
        nodes=[A, B],
        edges=[
            mdf.Edge(
                id="A->B",
                sender="A",
                receiver="B",
                sender_port="A_output",
                receiver_port="B_input",
            )
        ],
        conditions=mdf.ConditionSet(
            termination={
                # include JustRan condition so that
                # ENVIRONMENT_STATE_UPDATE isn't terminated immediately
                # after A reaches threshold when it is the condition's
                # dependency since we are testing B's output. This
                # ensures B will have its input and output updated
                graph_scheduler.TimeScale.ENVIRONMENT_STATE_UPDATE: mdf.Condition(
                    type="And",
                    dependencies=[
                        mdf.Condition(
                            type="Threshold",
                            dependency=threshold_dependency,
                            parameter=threshold_parameter,
                            threshold=5,
                            comparator=">=",
                            indices=threshold_indices,
                        ),
                        mdf.Condition(type="JustRan", dependency="B"),
                    ],
                )
            },
        ),
    )

    eg = EvaluableGraph(m.graphs[0])
    eg.evaluate(initializer={"A_input": 0})

    assert np.array_equal(
        eg.enodes["B"].evaluable_outputs["B_output"].curr_value, result
    )



import pytest
import numpy
from modeci_mdf.mdf import (
    Parameter,
    ParameterCondition,
)
from modeci_mdf.execution_engine import (
    EvaluableParameter,
)
from modelspec.utils import _params_info


@pytest.mark.parametrize(
    "value, parameters, result, cond_test, cond_value, default_initial_value",
    [
        ("3+4", {}, 7, None, None, None),
        ("3+x", {"x": 3}, 6.0, None, None, None),
        # ("p1+1", {"x": 3}, 3, None, None, None),
        ("[x,4]", {"x": 3}, [3, 4], None, None, None),
        ([1, 2], {"x": 3}, [1, 2], None, None, None),
        # ('a*b', {"a": 3, 'b':[1, 2]}, [3, 6]),
        (
            "a+b",
            {"a": numpy.array([1, 2]), "b": numpy.array([1, 2])},
            [2, 4],
            None,
            None,
            None,
        ),
        (numpy.array([1, 2]), {"x": 2}, numpy.array([1, 2]), None, None, None),
        (
            "x*y",
            {"x": 2, "y": numpy.array([1, 2])},
            numpy.array([2, 4]),
            None,
            None,
            None,
        ),
        (
            "x*y",
            {"x": numpy.array([1, 2]), "y": numpy.array([1, 2])},
            numpy.array([1, 4]),
            None,
            None,
            None,
        ),
        ### Conditions
        (0, {"x": 3}, 1, "x<4", 1, None),
        (0.6, {"x": 3}, 0.6, "x>=4", 1.6, None),
        ("y+z", {"x": 3, "y": 2, "z": 3}, 3, "x>y", "z", None),
        (
            numpy.array([2, 2]),
            {"x": numpy.array([1, 0]), "y": numpy.array([0, 1])},
            numpy.array([3, 2]),
            "x>y",
            [3, 3],
            None,
        ),
        # ([0,0], {"x": 3}, 1, 'x<4', 1, None),
        # (0, {}, 1, 'p1>2', 1, None),
        # ("x", {"x": 'abc'}, 'abc'),
    ],
)
def test_evaluable_parameter(
    value, parameters, result, cond_test, cond_value, default_initial_value
):

    p = Parameter(id="p1", value=value)
    if default_initial_value:
        p.default_initial_value = default_initial_value
    if cond_test:
        pc = ParameterCondition(id="pc", test=cond_test, value=cond_value)
        p.conditions.append(pc)

    print(f"---\nEvaluating the {p.summary()} with {parameters}...")

    ep = EvaluableParameter(p, verbose=True)
    evl = ep.evaluate(parameters)

    print(f"Evaluated {value} with {_params_info(parameters)} to be: {evl}")

    assert numpy.all(evl == result)
    assert numpy.all(
        ep.get_current_value(parameters) == result
    )  # should still remain the same, i.e. not evaluated again.



"""
Runs and tests everything in the examples folder.
"""

import pytest
import glob
import runpy
import os
import sys
import copy

from distutils.dir_util import copy_tree
from pathlib import Path

# Set this to True if you want for example scripts to be run from their current location in the repo rather than
# copied to a pytest temp directory. WARNING: setting this to True will cause files that these scripts generate to be
# re-generated.
run_example_scripts_in_repo = True

example_scripts = glob.glob("examples/**/*.py", recursive=True)
example_mdf_scripts = {
    Path(f) for f in glob.glob("examples/MDF/**/*.py", recursive=True)
}
example_pnl_scripts = {
    Path(f) for f in glob.glob("examples/PsyNeuLink/**/*.py", recursive=True)
}
example_exclusion_strings = [
    ".reconstructed.py",
    "generate_json_and_scripts.py",
    "pytorch_ddm.py",
]

# Filter any excluded example scripts
example_scripts = [
    script
    for script in example_scripts
    if all(e not in script for e in example_exclusion_strings)
]


@pytest.fixture(scope="session")
def example_tmp_dir(tmpdir_factory):
    """
    Create a temporary directory to run all example scripts from.
    This is a session scoped fixture so it is shared for all tests.
    """
    tmpdir = tmpdir_factory.mktemp("examples", numbered=False)
    copy_tree("examples", tmpdir.strpath)
    return tmpdir


@pytest.fixture(autouse=True)
def chdir_back_to_root(mocker):
    """
    This fixture sets up and tears down state before each example is run. Certain examples
    require that they are run from the local directory in which they reside. This changes
    directory and adds the local directory to sys.path. It reverses this after the test
    finishes.
    """

    # Get the current directory before running the test
    cwd = os.getcwd()

    # Some of the scripts do plots. Lets patch matplotlib plot so tests don't hang
    mocker.patch("matplotlib.pyplot.show")
    mocker.patch("matplotlib.pyplot.figure")

    # Cache the path so we can reset it after the test, the examples/MDF tests require
    # setting the path
    old_sys_path = copy.copy(sys.path)

    yield

    sys.path = old_sys_path

    # We need chdir back to root of the repo
    os.chdir(cwd)


@pytest.mark.parametrize("script", example_scripts)
@pytest.mark.parametrize("additional_args", [["-run"]])
def test_example(script, example_tmp_dir, additional_args):
    """
    Run the examples/MDF
    """

    # Change directory to the the parent of the examples temp directory
    # so that we can find the absolute path to the script we are testing
    # within the temporary directory and not the root of the repo.
    os.chdir(str(Path(example_tmp_dir).parent))

    # Get the full path for the script
    full_script_path = os.path.abspath(script)

    # Some of the scripts in examples/MDF import from the local directory. So lets run from the scripts
    # local directory.
    dir_path = os.path.dirname(os.path.realpath(full_script_path))
    os.chdir(dir_path)

    # If this is one of the example/MDF scripts, we need to append example/MDF to sys.path
    if Path(script) in example_mdf_scripts:
        sys.path.append(dir_path)

    print(f"Running script {full_script_path} in working dir {os.getcwd()}")
    orig_argv = sys.argv
    sys.argv = [os.path.basename(full_script_path)] + additional_args
    runpy.run_path(os.path.basename(full_script_path), run_name="__main__")
    sys.argv = orig_argv



import numpy as np

from modeci_mdf.mdf import (
    Model,
    Graph,
    Node,
    OutputPort,
    Function,
    Condition,
    ConditionSet,
    Parameter,
    Edge,
    InputPort,
)

from modeci_mdf.utils import load_mdf

import os
import pytest


def test_model_init_kwargs():
    m = Model(
        id="Test",
        format="test_format",
        generating_application="test_application",
        metadata={"info": "test_metadata"},
    )
    assert m.format == "test_format"
    assert m.generating_application == "test_application"
    assert m.metadata == {"info": "test_metadata"}
    assert m.id == "Test"


def test_graph_init_kwargs():
    g = Graph(
        id="Test_Graph", parameters={"test_parameters": 1}, conditions=ConditionSet()
    )
    assert g.parameters == {"test_parameters": 1}
    assert str(g.conditions) == str(ConditionSet())


def test_Node_init_kwargs():
    n = Node(id="test_node")
    print(n)
    print(n.id)
    assert n.id == "test_node"


def test_Function_init_kwargs():
    f = Function(id="Test_Function", function="Test_function", args={"Test_arg": 1})
    assert f.function == "Test_function"
    assert f.args == {"Test_arg": 1}


# def test_Function_init_kwargs():
#     f = Function(
#         id="Test_Function", function={"Test_function": "Linear"}, args={"Test_arg": 1}
#     )
#     assert f.function == {"Test_function": "Linear"}
#     # assert f.args == {"Test_arg": 1}


def test_InputPort_init_kwargs():
    ip = InputPort(id="Test_InputPort", shape=(1, 2), type="Test_type")
    assert ip.shape == (1, 2)
    assert ip.type == "Test_type"


def test_OutputPort_init_kwargs():
    op = OutputPort(id="test_OutputPort", value="test_value")
    assert op.value == "test_value"


def test_Edge_init_kwargs():
    e = Edge(
        id="test_Edge",
        parameters={"test_parameters": 3},
        sender="test_sender",
        receiver="test_receiver",
        sender_port="test_sender_port",
        receiver_port="test_receiver_port",
    )
    assert e.parameters == {"test_parameters": 3}
    assert e.sender == "test_sender"
    assert e.receiver == "test_receiver"
    assert e.sender_port == "test_sender_port"
    assert e.receiver_port == "test_receiver_port"
    assert e.id == "test_Edge"


def test_ConditionSet_init_kwargs():
    CS = ConditionSet(
        node_specific={"test_node_specific": 1}, termination={"test_termination": 3}
    )
    assert CS.node_specific == {"test_node_specific": 1}
    assert CS.termination == {"test_termination": 3}


def test_Condition_init_kwargs():
    """Check the working of Condition"""
    C = Condition(type="test_type", n="test_n", dependency="test_dependency")
    assert C.type == "test_type"
    assert C.kwargs == {"n": "test_n", "dependency": "test_dependency"}


def test_Condition_init_kwargs():
    C = Condition(type="test_type", n="test_n", dependencies="test_dependencies")
    assert C.type == "test_type"
    assert C.kwargs == {"n": "test_n", "dependencies": "test_dependencies"}


def test_model_graph_to_json():
    """Check if dumping a model to a simple JSON string works."""

    mod_graph0 = Graph(
        id="Test", parameters={"speed": 4}, metadata={"info": "mdf_model"}
    )

    node = Node(id="N0", metadata={"info": "mdf_Node"})
    node.parameters.append(Parameter(id="rate", value=5))

    node1 = Node(id="test_node", metadata={"info": "mdf_Node2"})
    node1.parameters.append(Parameter(id="level", value=3))

    condition = Condition(type="Always")
    condition.metadata = {"info": "mdf_condition"}

    mod_graph0.conditions = ConditionSet(node_specific={node1.id: condition})

    mod_graph0.nodes.append(node)
    mod_graph0.nodes.append(node1)

    # Export to JSON and see if we can load back in
    import json

    d = json.loads(mod_graph0.to_json())


def test_no_input_ports_to_json(tmpdir):
    """Test the edge case of exporting a model to JSON when it has a node with no input ports"""

    mod = Model(id="ABCD")
    mod_graph = Graph(id="abcd_example")
    mod.graphs.append(mod_graph)

    input_node = Node(id="input0")
    input_node.parameters.append(Parameter(id="input_level", value=10.0))
    op1 = OutputPort(id="out_port")
    op1.value = "input_level"
    input_node.output_ports.append(op1)
    mod_graph.nodes.append(input_node)

    tmpfile = f"{tmpdir}/test.json"
    mod.to_json_file(tmpfile)

    lmod = load_mdf(tmpfile)

    assert lmod.graphs[0].nodes[0].parameters[0].value == 10.0


def test_include_metadata_to_json(tmpdir):
    """
    Test for serialization
    """

    mod = Model(id="ABCD", metadata={"info": "model_test"})
    mod_graph = Graph(
        id="abcd_example", metadata={"info": {"graph_test": {"environment_x": "xyz"}}}
    )
    mod.graphs.append(mod_graph)

    input_node = Node(id="input0", metadata={"color": ".8 0 .8"})
    input_node.parameters.append(Parameter(id="input_level", value=10.0))
    op1 = OutputPort(id="out_port", metadata={"info": "value at OutputPort"})
    op1.value = "input_level"
    input_node.output_ports.append(op1)
    mod_graph.nodes.append(input_node)

    tmpfile = f"{tmpdir}/test.json"
    mod.to_json_file(tmpfile)

    lmod = load_mdf(tmpfile)
    lmod.graphs[0].metadata

    assert lmod.graphs[0].metadata == {"info": {"graph_test": {"environment_x": "xyz"}}}
    assert lmod.graphs[0].get_node("input0").metadata == {"color": ".8 0 .8"}
    assert lmod.graphs[0].get_node("input0").output_ports[0].metadata == {
        "info": "value at OutputPort"
    }


def test_node_params_empty():
    """
    Test whether we get a TypeError when passing no id for a node
    """
    with pytest.raises(TypeError):
        Node()


def test_metadata_dict():
    """
    Test whether we get a serialization error when passing anything else from a dictionary
    """
    with pytest.raises(Exception):
        Graph(id="n0", metadata="info").to_json()


def test_param_args_empty_dict():
    """
    Test whether we don't a serialization error when passing empty dicts to Parameter args
    """
    Parameter(id="noargs", args={}).to_json()


def test_graph_inputs():
    r"""Test whether we can retrieve graph input node\ports via the inputs property."""
    mod = Model(id="ABCD")
    mod_graph = Graph(id="abcd_example")
    mod.graphs.append(mod_graph)

    input_node = Node(id="input0")
    input_node.parameters.append(Parameter(id="input_level", value=10.0))
    op1 = OutputPort(id="out_port")
    op1.value = "input_level"
    input_node.output_ports.append(op1)
    mod_graph.nodes.append(input_node)


def test_graph_inputs_none(simple_model_mdf):
    """Test that the simple model with no input ports used has no graph inputs"""
    assert len(simple_model_mdf.graphs[0].inputs) == 0


def test_graph_types(tmpdir):
    r"""
    Test whether types saved in parameters are the same after reloading
    """
    mod = Model(id="Test0")
    mod_graph = Graph(id="test_example")
    mod.graphs.append(mod_graph)
    node0 = Node(id="node0")
    mod_graph.nodes.append(node0)

    p_int = 2
    node0.parameters.append(Parameter(id="p_int", value=p_int))
    p_float = 2.0
    node0.parameters.append(Parameter(id="p_float", value=p_float))
    p_bool = False
    node0.parameters.append(Parameter(id="p_bool", value=p_bool))
    p_str = "p_int + p_float"
    node0.parameters.append(Parameter(id="p_str", value=p_str))
    p_str2 = "2"
    node0.parameters.append(Parameter(id="p_str2", value=p_str2))
    p_list = ["2", 2, "two"]
    node0.parameters.append(Parameter(id="p_list", value=p_list))

    p_dict = {"a": 3, "w": {"b": 3, "x": True, "y": [2, 2, 2, 2]}}
    node0.parameters.append(Parameter(id="p_dict", value=p_dict))

    p_dict_tuple = {"y": (4, 44)}  # will change to {'y': [4, 44]}
    node0.parameters.append(Parameter(id="p_dict_tuple", value=p_dict_tuple))

    print(mod)
    tmpfile = f"{tmpdir}/test.json"
    mod.to_json_file(tmpfile)
    mod_graph2 = load_mdf(tmpfile)
    print(f"Saved to {tmpfile}: {mod_graph2}")
    new_node0 = mod_graph2.graphs[0].nodes[0]

    for p in [p.id for p in node0.parameters]:
        print(
            "Testing {}, is {} = {}?".format(
                p, new_node0.get_parameter(p).value, eval(p)
            )
        )

        assert type(new_node0.get_parameter(p).value) == type(eval(p))
        # Type will be same for tuple containing dict, but tuple will have been converetd to dict...
        if not p == "p_dict_tuple":
            assert new_node0.get_parameter(p).value == eval(p)


def test_arrays(tmpdir):
    """Test whether arrays are serialized and deserialized properly."""

    mod = Model(id="Arrays")
    mod_graph = Graph(id="array_example")
    mod.graphs.append(mod_graph)

    input_node = Node(id="input_node")

    input_node.parameters.append(Parameter(id="input_level", value=[[1, 2.0], [3, 4]]))

    op1 = OutputPort(id="out_port", value="input_level")
    input_node.output_ports.append(op1)
    mod_graph.nodes.append(input_node)

    middle_node = Node(id="middle_node")
    middle_node.parameters.append(Parameter(id="slope", value=0.5))
    middle_node.parameters.append(
        Parameter(id="intercept", value=np.array([[0, 1.0], [2, 2]]))
    )

    ip1 = InputPort(id="input_port1")
    middle_node.input_ports.append(ip1)
    mod_graph.nodes.append(middle_node)

    f1 = Parameter(
        id="linear_1",
        function="linear",
        args={"variable0": ip1.id, "slope": "slope", "intercept": "intercept"},
    )
    middle_node.parameters.append(f1)

    middle_node.output_ports.append(OutputPort(id="output_1", value="linear_1"))

    e1 = Edge(
        id="input_edge",
        parameters={"weight": [[1, 0], [0, 1]]},
        sender=input_node.id,
        sender_port=op1.id,
        receiver=middle_node.id,
        receiver_port=ip1.id,
    )

    mod_graph.edges.append(e1)

    json_file = mod.to_json_file(os.path.join(tmpdir, f"{mod.id}.json"))
    yaml_file = mod.to_yaml_file(os.path.join(tmpdir, f"{mod.id}.yaml"))
    bson_file = mod.to_bson_file(os.path.join(tmpdir, f"{mod.id}.bson"))

    json_mod = load_mdf(json_file)
    yaml_mod = load_mdf(yaml_file)
    bson_mod = load_mdf(bson_file)

    # Check both the YAML, JSON and BSON loaded versions of the model
    for mod in [json_mod, yaml_mod, bson_mod]:
        assert (
            mod.graphs[0].edges[0].parameters["weight"] == np.array([[1, 0], [0, 1]])
        ).all()


def test_ndarray_json_metadata():
    model = Node(id="a", metadata={"b": np.array([0])})
    model.to_json()


def test_serialization_Function_metadata():
    model = Function(id="a", metadata={"b": 0}, value="0")
    new_model = model.from_dict(model.to_dict())

    assert new_model.metadata == model.metadata


@pytest.mark.parametrize(
    "function, args",
    [
        ("linear", {"slope": 1, "intercept": 1}),
        ({"linear": {"slope": 1, "intercept": 1}}, None),
    ],
)
def test_serialization_Function_function_args_formats(function, args):
    model = Function("a", function=function, args=args)
    new_model = model.from_dict(model.to_dict())

    assert new_model.function == "linear"
    assert new_model.args == {"slope": 1, "intercept": 1}


def test_serialization_nested_Conditions():
    model = Condition(
        type="All",
        kwargs={
            "args": [
                Condition(
                    type="Not",
                    kwargs={
                        "condition": Condition(
                            type="BeforeNCalls",
                            kwargs={
                                "dependency": "B",
                                "n": 5,
                                "time_scale": "TimeScale.ENVIRONMENT_STATE_UPDATE",
                            },
                        )
                    },
                )
            ]
        },
    )
    new_model = model.from_dict(model.to_dict())

    cond_not = new_model.kwargs["args"][0]
    assert isinstance(cond_not, Condition)

    cond_beforencalls = cond_not.kwargs["condition"]
    assert isinstance(cond_beforencalls, Condition)

    assert cond_beforencalls.kwargs["dependency"] == "B"
    assert cond_beforencalls.kwargs["n"] == 5
    assert (
        cond_beforencalls.kwargs["time_scale"] == "TimeScale.ENVIRONMENT_STATE_UPDATE"
    )

    assert new_model.to_dict() == model.to_dict()


if __name__ == "__main__":
    test_graph_types("/tmp")



import numpy as np

__all__ = ["drift_diffusion_integrator"]


def drift_diffusion_integrator(
    starting_point: float,
    non_decision_time: float,
    drift_rate,
    threshold: float,
    noise: float = 1.0,
    dt: float = 0.01,
) -> float:
    """
    Integrates the drift diffusion model for a single trial using and implementation of
    the using the Euler-Maruyama method. This is a proof of concept implementation and
    is not optimized for speed.

    Args:
        starting_point: The starting point of the particle
        non_decision_time: The non-decision time
        drift_rate: The deterministic drift rate of the particle
        threshold: The threshold to cross, the boundary is assumed to be symmetric.
        noise: The standard deviation of the noise
        dt: The time step to use for the integration

    Returns:
        The time it took to cross the threshold, with the sign indicating the direction
    """
    particle = starting_point

    # Integrate the Wiener process until it crosses the threshold
    t = 0
    while abs(particle) < threshold:
        particle = particle + np.random.normal(
            loc=drift_rate * dt, scale=noise * np.sqrt(dt)
        )
        t = t + 1

    # Return the time it took to cross the threshold, with the sign indicating the direction
    # Add the non-decision time to the RT
    return (
        non_decision_time + t * dt
        if particle > threshold
        else -non_decision_time - t * dt
    )
